from gql import Client
from gql.transport.aiohttp import AIOHTTPTransport
from noloco.project import Project
from noloco.requests import Command
from noloco.utils import options_without_data


CORE_BASE_URL = 'https://api.core.noloco.io'
PROJECT_BASE_URL = 'https://api.portals.noloco.io'


class Noloco:
    def __init__(
        self,
        account_api_key,
        portal_name,
        core_base_url=CORE_BASE_URL,
        project_base_url=PROJECT_BASE_URL
    ):
        """Initialises a Noloco client.

        Args:
            account_api_key: The Account API Key from your Integrations & API
                Keys settings page.
            portal_name: The name of your Noloco portal.
            core_base_url: The URL that the core API is hosted at. This is an
                optional parameter and if you are using the production API you
                should not provide it.
            project_base_url: The URL that the project API is hosted at. This
                is an optional parameter and if you are using the production
                API you should not provide it.

        Returns:
            A Noloco client.

        Raises:
            NolocoAccountApiKeyError: If your Account API Key is incorrect.
            NolocoProjectApiKeyError: If we cannot fetch you Project API Key.
            NolocoUnknownError: If we are not sure what went wrong.
        """
        # Build the account client that will be used to interact with the
        # project document.
        account_transport = AIOHTTPTransport(
            url=core_base_url, headers={'Authorization': account_api_key})
        account_client = Client(
            transport=account_transport,
            fetch_schema_from_transport=False)

        # Fetch the project document, lookup and validate the project API key
        # and cache the data types locally.
        self.__project = Project(account_client, project_base_url, portal_name)

    def create(self, data_type_name, options):
        """Creates a record in a Noloco collection.

        Args:
            data_type_name: The name of the data type the collection is for.
                For example 'user'.
            options: The record to create as the 'data' field and any other
                options; after creating the record in the Noloco collection,
                the created record will be returned along with it's top-level
                fields. If you would like to also return some relationship
                fields you can do them using options. For example:

                {
                    'data': {
                        'firstName': 'Jane',
                        'lastName': 'Doe',
                        'email': 'jane@noloco.io',
                        'company': {
                            'connect': {
                                id: 2
                            }
                        },
                        'profilePicture': [open file]
                    },
                    'include': {
                        'company': {
                            'include': {
                                'usersCollection': True
                            }
                        }
                    }
                }

        Returns:
            The record that was created in the Noloco collection.
        """
        return Command(self.__project) \
            .for_data_type(data_type_name) \
            .with_options(options_without_data(options)) \
            .mutate('create') \
            .value(options['data']) \
            .with_pagination_callback(self.findUnique) \
            .build() \
            .execute()

    def delete(self, data_type_name, id):
        """Deletes a record from a Noloco collection.

        Args:
            data_type_name: The name of the data type the collection is for.
                For example 'user'.
            id: The ID of the record to delete.

        Returns:
            None.
        """
        return Command(self.__project) \
            .for_data_type(data_type_name) \
            .mutate('delete') \
            .with_id_lookup(id) \
            .build() \
            .execute()

    def findMany(self, data_type_name, options={}):
        """Searches a Noloco collection for records matching the provided
        criteria.

        Args:
            data_type_name: The name of the data type the collection is for.
                For example 'user'.
            options: The configuration for the search. Any matching records
                will be returned along with their top-level fields. If you
                would like to also return some relationship fields you can do
                them using options. For example:

                {
                    'after': '...',
                    'before': '...',
                    'first': '...',
                    'include': {
                        'role': True
                    },
                    'order_by': {
                        'direction': 'ASC',
                        'field': 'createdAt'
                    }
                    'where': {
                        'id': {
                            'gt': 5
                        }
                    }
                }

        Returns:
            The result of querying the Noloco collection.
        """
        return Command(self.__project) \
            .for_data_type(data_type_name) \
            .with_options(options) \
            .query('findMany') \
            .with_pagination_callback(self.findMany) \
            .build() \
            .execute()

    def findUnique(self, data_type_name, options):
        """Fetches a record from a Noloco collection that you identify by any
        of its unique fields.

        Args:
            data_type_name: The name of the data type you want to fetch. For
                example 'user'.
            options: The configuration for the lookup. The matching record will
                be returned along with its top-level fields. If you would like
                to also return some relationship fields you can do them using
                options. For example:

                {
                    'include': {
                        'company': True,
                        'role': True
                    }
                    'where': {
                        'id': {
                            'equals': 2
                        }
                    }
                }

        Returns:
            The result of looking up the Noloco record.
        """
        return Command(self.__project) \
            .for_data_type(data_type_name) \
            .with_options(options) \
            .query('findUnique') \
            .with_unique_lookup() \
            .with_pagination_callback(self.findUnique) \
            .build() \
            .execute()

    def update(self, data_type_name, id, options):
        """Updates a record in a collection.

        Args:
            data_type_name: The name of the data type the collection is for.
                For example 'user'.
            id: the ID of the record to update.
            options: The record to update as the 'data' field and any other
                options; after creating the record in the Noloco collection,
                the created record will be returned along with it's top-level
                fields. If you would like to also return some relationship
                fields you can do them using options. For example:

                {
                    'data': {
                        'firstName': 'Jane',
                        'lastName': 'Doe',
                        'email': 'jane@noloco.io',
                        'company': {
                            'connect': {
                                id: 2
                            }
                        },
                        'profilePicture': [open file]
                    },
                    'include': {
                        'role': True
                    }
                }

        Returns:
            The result of updating the Noloco record.
        """
        return Command(self.__project) \
            .for_data_type(data_type_name) \
            .with_options(options_without_data(options)) \
            .mutate('update') \
            .with_id_lookup(id) \
            .value(options['data']) \
            .with_pagination_callback(self.findUnique) \
            .build() \
            .execute()



from noloco.exceptions import (
    NolocoFieldNotFoundError,
    NolocoInvalidMultiFieldConnectionError,
    NolocoInvalidSingleFieldConnectionError,
    NolocoUnknownError)
from noloco.fields import DataTypeFieldsBuilder
from noloco.utils import (
    build_operation_args,
    gql_type,
    with_required,
    is_multi_relationship,
    is_reverse_multi_relationship, pascal_case,
    with_required)
from pydash import (
    find,
    get
)


DATA_TYPE_MUTATION = '''mutation{mutation_args} {{
  {mutation_fragment}
}}'''


class MutationBuilder:
    def __init__(self):
        self.fields_builder = DataTypeFieldsBuilder()

    def build_data_type_mutation_args(self, data_type, data_types, args):
        mutation_args = {}

        for arg_name, arg_value in args.items():
            data_type_fields = [
                field
                for field
                in data_type['fields']
                if field['name'] == arg_name]
            data_type_field = find(data_type_fields)

            if data_type_field is not None:
                is_required = data_type_field['required']

                # The field is either a top-level or relationship field on the
                # data type.
                if data_type_field['relationship'] is None and \
                        data_type_field['type'] != 'file':
                    # This is a top-level field, so map the arg onto a
                    # primitive.
                    mutation_args[arg_name] = {
                        'type': gql_type(
                            data_type,
                            data_type_field,
                            is_required),
                        'value': arg_value}
                elif get(arg_value, 'connect') is not None:
                    # This is a relationship field, so map the arg onto an Id
                    # arg.
                    # TODO - stronger validation and error handling.
                    # TODO - consider supporting connecting on other fields.

                    connection_value = arg_value['connect']
                    if is_multi_relationship(data_type_field['relationship']):

                        if not isinstance(connection_value, list):
                            raise NolocoInvalidMultiFieldConnectionError(
                                arg_name)

                        mutation_args[arg_name + 'Id'] = {
                            'type': with_required('[ID!]', is_required),
                            'value': [connection['id'] for connection in connection_value]
                        }
                    else:
                        if not isinstance(connection_value, dict):
                            raise NolocoInvalidSingleFieldConnectionError(
                                arg_name)

                        mutation_args[arg_name + 'Id'] = {
                            'type': with_required('ID', is_required),
                            'value': connection_value['id']
                        }
                elif data_type_field['type'] == 'file':
                    # This is a file upload field, so map the arg onto an
                    # Upload arg, although do not open the file yet.
                    # TODO - stronger validation and error handling.
                    if is_multi_relationship(data_type_field['relationship']):
                        mutation_args[arg_name] = {
                            'type': with_required('[Upload!]', is_required),
                            'value': arg_value
                        }
                    else:
                        mutation_args[arg_name] = {
                            'type': with_required('Upload', is_required),
                            'value': arg_value
                        }
                else:
                    raise NolocoUnknownError()
            else:
                # The field is a reverse relationship field to the data type or
                # doesn't exist.
                for related_data_type in data_types:
                    for field in related_data_type['fields']:
                        if field['type'] == data_type['type'] and \
                                field['reverseName'] == arg_name:
                            related_field = field

                if related_field is not None:
                    # This is a reverse relationship field and can be
                    # connected.
                    if is_reverse_multi_relationship(related_field['relationship']):
                        connection_value = arg_value['connect']

                        if not isinstance(connection_value, list):
                            raise NolocoInvalidMultiFieldConnectionError(
                                arg_name)

                        mutation_args[arg_name + 'Id'] = {
                            'type': with_required('[ID!]', related_field['required']),
                            'value': [connection['id'] for connection in connection_value]
                        }
                    else:
                        connection_value = arg_value['connect']

                        if not isinstance(connection_value, dict):
                            raise NolocoInvalidSingleFieldConnectionError(
                                arg_name)

                        mutation_args[arg_name + 'Id'] = {
                            'type': with_required('ID', related_field['required']),
                            'value': connection_value['id']
                        }
                else:
                    # This field doesn't exist on the type.
                    raise NolocoFieldNotFoundError(arg_name)

        return mutation_args

    def build_data_type_mutation(
            self,
            mutation,
            data_type,
            data_types,
            options,
            flattened_options):
        mutation_args = build_operation_args(flattened_options)

        mutation_fragment = self.fields_builder.build_fields(
            mutation + pascal_case(data_type['name']),
            data_type,
            data_types,
            options)

        return DATA_TYPE_MUTATION.format(
            mutation_args=mutation_args,
            mutation_fragment=mutation_fragment)



from noloco.constants import (
    MANY_TO_ONE,
    ONE_TO_ONE)
from noloco.utils import (
    build_data_type_args,
    find_field_by_name,
    find_relationship_data_type,
    is_multi_relationship)


DATA_TYPE_FIELDS = '''{data_type_name}{data_type_args}
{{
  {data_type_schema}
}}'''


DATA_TYPE_COLLECTION_FIELDS = '''{data_type_name}{data_type_args} {{
    totalCount
    edges {{
      node {{
        {data_type_schema}
      }}
    }}
    pageInfo {{
      hasPreviousPage
      hasNextPage
      startCursor
      endCursor
    }}
    __typename
  }}
'''


FILE_FIELDS = '''id uuid fileType url name'''


FILE_CONNECTION_FIELDS = '''totalCount
edges {{
  node {{
    {file_query}
  }}
}}
pageInfo {{
  hasPreviousPage
  hasNextPage
  startCursor
  endCursor
}}
__typename'''.format(file_query=FILE_FIELDS)


class DataTypeFieldsBuilder:
    def __build_related_fields(
            self,
            data_type_name,
            data_type_full_name,
            fields,
            include,
            data_types):
        related_fields = []

        for relationship_name, ignore_children in include.items():
            relationship_data_type = find_relationship_data_type(
                relationship_name,
                data_type_name,
                fields,
                data_types)

            # For example if include={'usersCompleted': True} was passed in,
            # we will not include any relationships from the User data type.
            # when including the usersCompleted related field. However, if
            # include={'usersCompleted': {'include': {'company': True}}} was
            # passed in, we would recursively include the company relationship
            # against any returned users.
            if ignore_children is True:
                response = {}
            else:
                response = ignore_children

            is_collection = relationship_data_type['is_collection']

            relationship_schema = self.build_fields(
                relationship_name,
                relationship_data_type['data_type'],
                data_types,
                response,
                data_type_path=data_type_full_name + '_',
                is_collection=is_collection)

            related_fields.append(relationship_schema)

        return related_fields

    def __build_file_fields(self, files):
        file_fields = []

        for file in files:
            if file['relationship'] == ONE_TO_ONE or \
                    file['relationship'] == MANY_TO_ONE:
                file_fields.append(
                    file['name'] + ' { ' + FILE_FIELDS + ' }')
            else:
                file_fields.append(
                    file['name'] + ' { ' + FILE_CONNECTION_FIELDS + ' }')

        return file_fields

    def __build_data_type_schema(
            self,
            data_type_full_name,
            data_type,
            data_types,
            include):
        # All non-relationship fields on the data type are automatically
        # included in the requested schema.
        primary_field_schema = [
            field['name']
            for field
            in data_type['fields']
            if field['relationship'] is None]

        # Only specified relationship types are included in the requested
        # schema. This principle is applied recursively so if we include a
        # relationship field, we only include relationships from that field if
        # they are also specified.
        related_field_schema = self.__build_related_fields(
            data_type['name'],
            data_type_full_name,
            data_type['fields'],
            include,
            data_types)

        # All file relationship fields on the data type are automatically
        # included in the requested schema.
        file_field_schema = self.__build_file_fields(
            field for field in data_type['fields'] if field['type'] == 'file')

        all_field_names = primary_field_schema + \
            related_field_schema + \
            file_field_schema + ['__typename']

        return '\n'.join(all_field_names)

    def build_fields(
            self,
            data_type_name,
            data_type,
            data_types,
            response,
            data_type_path='',
            is_collection=False):
        data_type_full_name = data_type_path + data_type_name

        # Each top level key on the response object gets fully qualified by the
        # nesting path and mapped ont
        args = {}
        include = {}
        for arg_name, arg_value in response.items():
            if arg_name == 'include':
                include = arg_value
            else:
                args[data_type_full_name + '_' + arg_name] = arg_value
        data_type_args = build_data_type_args(args)

        data_type_schema = self.__build_data_type_schema(
            data_type_full_name, data_type, data_types, include)

        if is_collection:
            base_fragment = DATA_TYPE_COLLECTION_FIELDS
        else:
            base_fragment = DATA_TYPE_FIELDS

        return base_fragment.format(
            data_type_name=data_type_name,
            data_type_args=data_type_args,
            data_type_schema=data_type_schema)



from pydash import (
    get,
    set_)
from re import sub


class Result(dict):
    def __init__(
            self,
            data_type_name,
            result_name,
            result,
            options_path,
            options,
            client):
        for key, value in result.items():
            if options_path == '':
                next_options_path = '.'.join(['include', key])
            else:
                next_options_path = '.'.join([options_path, 'include', key])

            if type(value) is dict:
                if get(value, 'edges') is not None:
                    # Wrap a collection in a CollectionResult to give it
                    # pagination helpers.
                    result[key] = CollectionResult(
                        data_type_name,
                        result_name,
                        value,
                        next_options_path,
                        options,
                        client)
                else:
                    # Otherwise traverse each field in the result and
                    # recursively wrap any that are collections.
                    result[key] = Result(
                        data_type_name,
                        result_name,
                        value,
                        next_options_path,
                        options,
                        client)
            else:
                result[key] = value

        dict.__init__(self, result)

    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value

    @staticmethod
    def build(data_type_name, result_name, result, options, client):
        result = Result(
            data_type_name,
            result_name,
            result,
            '',
            {'include': {result_name: options}},
            client)

        return result[result_name]


class CollectionResult:
    def __init__(
            self,
            data_type_name,
            result_name,
            result,
            result_path,
            options,
            client):
        self.__client = client
        self.__data_type_name = data_type_name
        self.__options = options
        self.__result_path = result_path
        self.__page_info = result['pageInfo']
        self.__result_name = result_name

        self.total_count = result['totalCount']
        self.has_previous_page = result['pageInfo']['hasPreviousPage']
        self.has_next_page = result['pageInfo']['hasNextPage']
        self.data = []
        for index, edge in enumerate(result['edges']):
            self.data.append(
                Result(
                    data_type_name,
                    result_name,
                    edge['node'],
                    result_path + f'[{index}]',
                    options,
                    client))

    def __str__(self):
        total_count_str = f'\'total_count\': {str(self.total_count)}'
        has_previous_page_str = \
            f'\'has_previous_page\': {str(self.has_previous_page)}'
        has_next_page_str = f'\'has_next_page\': {str(self.has_next_page)}'
        data_str = f'\'data\': {str(self.data)}'

        properties_str = ', '.join([
            total_count_str,
            has_previous_page_str,
            has_next_page_str,
            data_str])

        return f'{{{properties_str}}}'

    def __options_path(self):
        return sub('\[[0-9]+\]', '', self.__result_path)

    def __page_path(self):
        unwrapped_path = sub(
            '^include\.' + self.__result_name,
            '',
            self.__result_path)
        return unwrapped_path.replace('include.', '').replace('[', 'data[')

    def __page(self, paged_options):
        # Overwrite the options that applied to this collection with the new
        # options and then unwrap these to remove the artificial
        # 'include': { data_type_name: ... } that we added when building the
        # result.
        client_options = self.__options
        set_(client_options, self.__options_path(), paged_options)
        client_options = get(
            client_options,
            f'include.{self.__result_name}')

        # Re-run the equivalent query that produced this page originally with
        # the new options.
        result = self.__client(
            self.__data_type_name,
            client_options)

        # Determine where in the result the page is returned and hoist it up to
        # return it back.
        page_path = self.__page_path()
        if page_path == '':
            return result
        else:
            return get(result, page_path)

    def previous_page(self):
        if not self.__page_info['hasPreviousPage']:
            return None
        else:
            # Fetch the options that applied to this collection, remove the
            # 'after' parameter if it exists and set the 'before' parameter to
            # the start cursor of the current page.
            options = get(self.__options, self.__options_path())
            options.pop('after', None)
            options['before'] = self.__page_info['startCursor']

            return self.__page(options)

    def next_page(self):
        if not self.__page_info['hasNextPage']:
            return None
        else:
            # Fetch the options that applied to this collection, remove the
            # 'before' parameter if it exists and set the 'after' parameter to
            # the end cursor of the current page.
            options = get(self.__options, self.__options_path())
            options.pop('before', None)
            options['after'] = self.__page_info['endCursor']

            return self.__page(options)



from noloco.fields import DataTypeFieldsBuilder
from noloco.utils import build_operation_args


PROJECT_DOCUMENT_QUERY = '''query ($projectId: String!) {
  project(projectId: $projectId) {
    id
    name
    apiKeys {
      user
      project
      __typename
    }
    dataTypes {
      id
      name
      display
      internal
      fields {
        id
        name
        display
        type
        unique
        required
        relationship
        reverseDisplayName
        reverseName
        options {
          id
          name
          display
          __typename
        }
        __typename
      }
      __typename
    }
    __typename
  }
}'''


VALIDATE_API_KEYS_QUERY = '''query ($projectToken: String!) {
  validateApiKeys(projectToken: $projectToken) {
    user {
      id
      email
      __typename
    }
    projectName
    __typename
  }
}'''


DATA_TYPE_QUERY = '''query{query_args} {{
  {data_type_fragment}
}}'''


DATA_TYPE_COLLECTION_QUERY = '''query{query_args} {{
  {data_type_fragment}
}}'''


class QueryBuilder:
    def __init__(self):
        self.fields_builder = DataTypeFieldsBuilder()

    def build_data_type_query(
            self,
            query_type,
            result_name,
            data_type,
            data_types,
            options,
            flattened_options):
        is_collection = query_type == 'findMany'  # TODO - clean this up
        query_args = build_operation_args(flattened_options)

        query_fragment = self.fields_builder.build_fields(
            result_name,
            data_type,
            data_types,
            options,
            '',
            is_collection)

        if is_collection:
            query_template = DATA_TYPE_COLLECTION_QUERY
        else:
            query_template = DATA_TYPE_QUERY

        query = query_template.format(
            query_args=query_args,
            data_type_fragment=query_fragment)

        return query



from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.exceptions import TransportQueryError
from noloco.exceptions import (
    NolocoAccountApiKeyError,
    NolocoProjectApiKeyError,
    NolocoUnknownError)
from noloco.queries import (
    PROJECT_DOCUMENT_QUERY,
    VALIDATE_API_KEYS_QUERY)
from pydash import get


class Project:
    def __init__(self, account_client, base_url, name):
        self.__account_client = account_client
        self.__base_url = base_url
        self.__project_name = name

        self.refresh()

    def refresh(self):
        # Try to validate the account API key and fetch the project API key.
        try:
            project_document_query_result = self.__account_client.execute(
                gql(PROJECT_DOCUMENT_QUERY),
                variable_values={'projectId': self.__project_name})
            project_api_key = get(
                project_document_query_result,
                'project.apiKeys.project')
            self.data_types = get(
                project_document_query_result, 'project.dataTypes')
        except TransportQueryError as err:
            raise NolocoAccountApiKeyError(self.__project_name, err)
        except Exception as err:
            raise NolocoUnknownError(err)

        # Try to validate the project API key.
        try:
            self.__account_client.execute(
                gql(VALIDATE_API_KEYS_QUERY),
                variable_values={'projectToken': project_api_key})
        except TransportQueryError as err:
            raise NolocoProjectApiKeyError(self.__project_name, err)
        except Exception as err:
            raise NolocoUnknownError(err)

        # Build the project client that will be used to interact with
        # collections.
        project_transport = AIOHTTPTransport(
            url=f'{self.__base_url}/data/{self.__project_name}',
            headers={'Authorization': project_api_key})
        self.client = Client(
            transport=project_transport,
            fetch_schema_from_transport=False)



from setuptools import (
    setup,
    find_packages)


setup(
    name='noloco',
    version='0.6.1',
    description='CRUD operations for Noloco Collections',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Noloco',
    author_email='engineering@noloco.io',
    license='MIT',
    packages=find_packages(exclude=('docs', 'tests')),
    include_package_data=True,
    install_requires=['gql[all]', 'pydash']
)



from noloco.constants import (
    BOOLEAN,
    COLLECTION,
    DATE,
    DECIMAL,
    DURATION,
    INTEGER,
    TEXT,
    MANY_TO_MANY,
    MANY_TO_ONE,
    MULTIPLE_OPTION,
    ONE_TO_MANY,
    ONE_TO_ONE,
    SINGLE_OPTION,
    TEXT)
from noloco.exceptions import (
    NolocoDataTypeNotFoundError,
    NolocoQueryNotSupportedError)
from pydash import (
    find,
    get,
    camel_case)


def pascal_case(str):
    camel = camel_case(str)
    return camel[:1].upper() + camel[1:]


def annotate_collection_args(data_type, data_types, args):
    annotated_args = {}

    # Process top-level supported parameters.
    if get(args, 'after') is not None:
        annotated_args['after'] = {'type': 'String', 'value': args['after']}
    if get(args, 'after') is None and get(args, 'before') is not None:
        annotated_args['before'] = {'type': 'String', 'value': args['before']}
    if get(args, 'first') is not None:
        annotated_args['first'] = {'type': 'Int', 'value': args['first']}
    if get(args, 'order_by') is not None:
        annotated_args['orderBy'] = {
            'type': 'OrderBy',
            'value': args['order_by']
        }
    if get(args, 'where') is not None:
        whereType = pascal_case(data_type['name']) + 'WhereInput'
        annotated_args['where'] = {'type': whereType, 'value': args['where']}

    # Recursively process nested supported parameters.
    if get(args, 'include') is not None:
        annotated_args['include'] = {}

        for nested_data_type_name, nested_args in args['include'].items():
            if nested_args is not True:
                relationship_data_type = find_relationship_data_type(
                    nested_data_type_name,
                    data_type['name'],
                    data_type['fields'],
                    data_types)['data_type']
                annotated_args['include'][nested_data_type_name] = \
                    annotate_collection_args(
                        relationship_data_type,
                        data_types,
                        nested_args)
            else:
                annotated_args['include'][nested_data_type_name] = True

    return annotated_args


def build_operation_arg(arg_name, arg_value):
    arg_type = arg_value['type']
    return f'${arg_name}: {arg_type}'


def build_operation_args(args):
    operation_arg_list = ', '.join([build_operation_arg(
        arg_name, arg_value) for arg_name, arg_value in args.items()])

    if operation_arg_list:
        return f'({operation_arg_list})'
    else:
        return ''


def build_data_type_arg(arg_full_name):
    arg_name = arg_full_name.split('_')[-1]
    return f'{arg_name}: ${arg_full_name}'


def build_data_type_args(args):
    data_type_arg_list = ', '.join(
        [build_data_type_arg(arg_name) for arg_name in args.keys()])
    if data_type_arg_list:
        return f'({data_type_arg_list})'
    else:
        return ''


def change_where_to_lookup(data_type, options):
    where = options.pop('where')['value']
    lookup_key = list(where.keys())[0]

    lookup_field = find_field_by_name(lookup_key, data_type['fields'])
    lookup_type = gql_type(data_type, lookup_field)
    lookup_value = where[lookup_key]['equals']

    if lookup_key == 'id':
        lookup_type = 'ID'

    options[lookup_key] = {'type': lookup_type, 'value': lookup_value}

    return options


def find_data_type_by_name(data_type_name, data_types):
    data_type = find(
        data_types,
        lambda project_data_type: project_data_type['name'] == data_type_name)

    if data_type is None:
        raise NolocoDataTypeNotFoundError(data_type_name)
    else:
        return data_type


def find_field_by_name(field_name, fields):
    field = find(
        fields,
        lambda data_type_field: data_type_field['name'] == field_name)

    return field


def find_relationship_data_type(
        relationship_name,
        data_type_name,
        data_type_fields,
        data_types):
    relationship_field = find_field_by_name(
        relationship_name,
        data_type_fields)

    if relationship_field is not None:
        # If the relationship field exists on the parent data type then
        # this is a forward relationship and we can simply look up the
        # relationship data type by the corresponding field type.
        return {
            'data_type': find_data_type_by_name(
                relationship_field['type'], data_types),
            'is_collection': is_multi_relationship(relationship_field['relationship'])
        }
    else:
        # If there isn't a corresponding relationship field on the
        # parent data type then this is a reverse relationship and we
        # have to search for the relationship data type.
        for candidate_data_type in data_types:
            for field in candidate_data_type['fields']:
                # We are looking for any fields whose type matches up with the
                # name of the parent data type that the reverse relationship is
                # from.
                if field['type'] == data_type_name:
                    # If the reverse name is populated and matches the parent
                    # type then we have found the data type the relationship
                    # goes to.
                    is_collection = field['relationship'] == MANY_TO_MANY or \
                        field['relationship'] == MANY_TO_ONE
                    if reverse_name_matches_relationship_name(
                            field, is_collection, relationship_name):
                        return {
                            'data_type': candidate_data_type,
                            'is_collection': is_collection
                        }
                    # If the reverse name is not populated but this field is a
                    # one-to-one relationship and the data type name matches
                    # the relationship name then we have found the data type
                    # the relationship goes to.
                    elif field['relationship'] == ONE_TO_ONE:
                        if candidate_data_type['name'] == relationship_name:
                            return {
                                'data_type': candidate_data_type,
                                'is_collection': False
                            }


def flatten_args(data_type_name, args):
    flattened_args = {}

    for arg_name, arg_value in args.items():
        if arg_name != 'include':
            flattened_args[data_type_name + '_' + arg_name] = arg_value
        else:
            for nested_data_type_name, nested_args in args['include'].items():
                if nested_args is not True:
                    flattened_nested_args = flatten_args(
                        data_type_name + '_' + nested_data_type_name,
                        nested_args)

                    for flattened_nested_arg in flattened_nested_args.keys():
                        flattened_args[flattened_nested_arg] = \
                            flattened_nested_args[flattened_nested_arg]

    return flattened_args


def gql_args(args):
    variables = {}

    for arg in args:
        variables[arg] = args[arg]['value']

    return variables


def gql_type(data_type, data_type_field, is_required=False):
    field_type = data_type_field['type']

    if field_type == TEXT:
        return with_required('String', is_required)
    elif field_type == DATE:
        return with_required('DateTime', is_required)
    elif field_type == INTEGER:
        return with_required('Int', is_required)
    elif field_type == DECIMAL:
        return with_required('Float', is_required)
    elif field_type == DURATION:
        return with_required('Duration', is_required)
    elif field_type == BOOLEAN:
        return with_required('Boolean', is_required)
    elif field_type == SINGLE_OPTION:
        enum_prefix = pascal_case(data_type['name'])
        enum_suffix = pascal_case(data_type_field['name'])
        return with_required(f'{enum_prefix}{enum_suffix}', is_required)
    elif field_type == MULTIPLE_OPTION:
        enum_prefix = pascal_case(data_type['name'])
        enum_suffix = pascal_case(data_type_field['name'])
        return with_required(f'[{enum_prefix}{enum_suffix}!]', is_required)


def has_files(args):
    for arg_value in args.values():
        if 'Upload' in arg_value['type']:
            return True

    return False


def options_without_data(options):
    new_options = {}

    for option in options.keys():
        if option != 'data':
            new_options[option] = options[option]

    return new_options


def result_name_suffix(query):
    if query == 'findMany':
        return 'Collection'
    elif query == 'findUnique':
        return ''
    else:
        raise NolocoQueryNotSupportedError(query)


def is_multi_relationship(relationship):
    return relationship == ONE_TO_MANY or relationship == MANY_TO_MANY


def is_reverse_multi_relationship(relationship):
    return relationship == MANY_TO_ONE or relationship == MANY_TO_MANY


def reverse_name_matches_relationship_name(
        field,
        is_collection,
        relationship_name):
    reverse_name = field['reverseName']

    if reverse_name is None or reverse_name == '':
        return False
    elif is_collection:
        reverse_collection_name = field['reverseName'] + COLLECTION
        return reverse_collection_name == relationship_name
    else:
        return reverse_name == relationship_name


def with_required(gql_type, is_required):
    if is_required:
        return f'{gql_type}!'
    else:
        return gql_type



class NolocoAccountApiKeyError(Exception):
    def __init__(self, project_name, error):
        super().__init__(
            'Your account API key did not authenticate for portal '
            f'{project_name}')
        self.project_name = project_name
        self.error = error


class NolocoDataTypeNotFoundError(Exception):
    def __init__(self, data_type_name):
        super().__init__(
            f'Could not find data type {data_type_name} in your project.')


class NolocoFieldNotFoundError(Exception):
    def __init__(self, field_name):
        super().__init__(
            f'Could not find field {field_name}.')


class NolocoFieldNotUniqueError(Exception):
    def __init__(self, data_type_name, field_name):
        super().__init__(
            f'Field {field_name} in data type {data_type_name} is not unique.')


class NolocoInvalidMultiFieldConnectionError(Exception):
    def __init__(self, argument_name):
        super().__init__(
            f'Value for argument "{argument_name}" must be a list of connections')


class NolocoInvalidSingleFieldConnectionError(Exception):
    def __init__(self, argument_name):
        super().__init__(
            f'Value for argument "{argument_name}" must be a single connection')


class NolocoProjectApiKeyError(Exception):
    def __init__(self, project_name, error):
        super().__init__(
            'We could not validate the API client we setup for portal '
            f'{project_name}')
        self.project_name = project_name
        self.error = error


class NolocoQueryNotSupportedError(Exception):
    def __init__(self, query_name):
        super().__init__(
            f'{query_name} queries are not supported.')


class NolocoUnknownError(Exception):
    def __init__(self, error):
        super().__init__('Something went wrong!')
        self.error = error



from gql import gql
from gql.transport.exceptions import TransportQueryError
from noloco.constants import GRAPHQL_VALIDATION_FAILED
from noloco.exceptions import (
    NolocoDataTypeNotFoundError,
    NolocoFieldNotFoundError)
from noloco.mutations import MutationBuilder
from noloco.queries import QueryBuilder
from noloco.results import Result
from noloco.utils import (
    annotate_collection_args,
    change_where_to_lookup,
    find_data_type_by_name,
    flatten_args,
    gql_args,
    has_files,
    result_name_suffix,
    pascal_case)
from pydash import (
    get)


class Command:
    def __init__(self, project):
        self.project = project
        self.__mutation_builder = MutationBuilder()
        self.__query_builder = QueryBuilder()

        self.data_type_name = None
        self.id_lookup = None
        self.id_lookup_type = None
        self.mutation = None
        self.new_value = None
        self.options = {}
        self.pagination_callback = None
        self.query_type = None
        self.result_name = None
        self.unique_lookup = None

    def for_data_type(self, data_type_name):
        self.data_type_name = data_type_name
        return self

    def with_id_lookup(self, id_lookup):
        self.id_lookup = id_lookup
        return self

    def with_options(self, options):
        self.options = options
        return self

    def with_pagination_callback(self, pagination_callback):
        self.pagination_callback = pagination_callback
        return self

    def value(self, new_value):
        self.new_value = new_value
        return self

    def with_unique_lookup(self):
        self.unique_lookup = True
        return self

    def mutate(self, mutation):
        self.mutation = mutation
        self.result_name = mutation + pascal_case(
            self.data_type_name)
        return self

    def query(self, query_type):
        self.query_type = query_type
        self.result_name = self.data_type_name + result_name_suffix(query_type)
        return self

    def build(self, retry=True):
        try:
            data_types = self.project.data_types
            data_type = find_data_type_by_name(self.data_type_name, data_types)

            typed_options = annotate_collection_args(
                data_type,
                data_types,
                self.options)

            if self.id_lookup is not None:
                typed_options['id'] = {'type': 'ID!', 'value': self.id_lookup}
            elif self.unique_lookup is not None:
                typed_options = change_where_to_lookup(
                    data_type,
                    typed_options)

            if self.new_value is not None:
                mutation_args = self \
                    .__mutation_builder \
                    .build_data_type_mutation_args(
                        data_type,
                        data_types,
                        self.new_value)
                upload_files = has_files(mutation_args)
                typed_options.update(mutation_args)
            else:
                upload_files = False

            flattened_options = flatten_args(self.result_name, typed_options)

            if self.mutation is not None:
                document = self.__mutation_builder.build_data_type_mutation(
                    self.mutation,
                    data_type,
                    data_types,
                    typed_options,
                    flattened_options)

            if self.query_type is not None:
                document = self.__query_builder.build_data_type_query(
                    self.query_type,
                    self.result_name,
                    data_type,
                    data_types,
                    typed_options,
                    flattened_options)

            return BuiltCommand(
                self,
                gql(document),
                gql_args(flattened_options),
                upload_files,
                self.pagination_callback)
        except (NolocoDataTypeNotFoundError, NolocoFieldNotFoundError):
            if retry:
                self.project.refresh()
                self.build(retry=False)
            else:
                raise


class BuiltCommand:
    def __init__(
            self,
            command,
            document,
            variable_values,
            upload_files,
            pagination_callback):
        self.__command = command
        self.__document = document
        self.__variable_values = variable_values
        self.__upload_files = upload_files
        self.__pagination_callback = pagination_callback

    def execute(self, retry=True):
        try:
            raw_result = self.__command.project.client.execute(
                self.__document,
                variable_values=self.__variable_values,
                upload_files=self.__upload_files)

            return Result.build(
                self.__command.data_type_name,
                self.__command.result_name,
                raw_result,
                self.__command.options,
                self.__pagination_callback)
        except TransportQueryError as err:
            if get(err, 'extensions.code') == GRAPHQL_VALIDATION_FAILED and \
                    retry:
                self.__command.project.refresh()
                return self.__command.build().execute(retry=False)
            else:
                raise
